/*
 * asmloader.c
 *
 * Linux Loader for the Palm OS.
 *
 *
 * Copyright (C) 2004 Vinayak R. Borkar
 * Author: Vinayak R. Borkar [vinayakb@users.sourceforge.net]
 *
 * Original version by Romain Goyet [r.goyet@gmail.com]
 *
 * 11-Sep-2005, Alex Osborne [bobofdoom@gmail.com]
 *   Updated to work with multiple handhelds.
 *
 *   Based on the work of:
 *   Nikolay Petukhov   - Palm Tungsten T2
 *   Vladimir Pouzanov  - Palm Tungsten T3
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <PceNativeCall.h>
#include <Standalone.h>
//#include <ByteOrderUtils.h>

#define EndianSwap32(n) (((((unsigned long) n) << 24) & 0xFF000000) |   ((((unsigned long) n) <<  8) & 0x00FF0000) |   ((((unsigned long) n) >>  8) & 0x0000FF00) |   ((((unsigned long) n) >> 24) & 0x000000FF))

#include "garux.h"
#include "pxa-regs.h"

STANDALONE_CODE_RESOURCE_ID (0);

void startKernel();
unsigned long get_rambase(unsigned long who);
unsigned long get_cpu(unsigned long who);
unsigned long get_ram_size(unsigned long who);

void init_xscale(unsigned long mach);
void init_omap(unsigned long mach);
void init_tt2(unsigned long mach);

void write_atags(unsigned long mach, unsigned long rambase, char *cmdline);


/**
 * Note: This is very hacky. We're relying on compiling with -O2 so that gcc
 * will use register to store mach & rambase.
 */
unsigned long bootLinux(const void *emulStateP, char *userData68KP, Call68KFuncType *call68KFuncP) {
  unsigned long register i;
  unsigned long register j;
  unsigned long register krnlSrc;
  char *cmdline;
  unsigned long register mach;
  unsigned long register rambase;
  
  mach =    EndianSwap32(*(unsigned long*)userData68KP);
  cmdline = (char*)(userData68KP+4);
  krnlSrc=(unsigned long)userData68KP+256+4;
  rambase = get_rambase(mach);
  
  // backup mach & rambase
  // and hope to hell the compiler doesn't use r12
  // it doesn't seem to for me
  //asm volatile ("mov r12, %0" : : "r" (mach));
  asm volatile ("mov r11, %0" : : "r" (rambase));

  // Disable FIQ, IRQ
  asm volatile ("mrs r0, cpsr");
  asm volatile ("orr r0, r0, #0xc0");
  asm volatile ("msr cpsr, r0");

  if (mach == OMAP_PALMTT2) {
    init_tt2(mach);
    // Set Page table entries, On palm tt2 page table place in sram.
    *((unsigned long*)0x20010400)=0x10000c1a;
    *((unsigned long*)0x20010440)=0x11000c1a;

  } else if (get_cpu(mach) == OMAP) {
    init_omap(mach);
    // Set Page table entries for 0x10000000 and 0x11000000 to identity.
    *((unsigned long *)0x4400)=0x10000c1a;
    *((unsigned long *)0x4440)=0x11000c1a;

  } else if (get_cpu(mach) == XSCALE) {
    init_xscale(mach);
    // Set Page table entries for 0xA0000000 and 0xA1000000 to identity.
    *((unsigned long *)0x6800)=0xA0000c1a;
    *((unsigned long *)0x6804)=0xA0100c1a;
    *((unsigned long *)0x6808)=0xA0200c1a;
    *((unsigned long *)0x6840)=0xA1000c1a;
    *((unsigned long *)0x6844)=0xA1100c1a;
    *((unsigned long *)0x6848)=0xA1200c1a;
  }

  // Let's flush Instruction and Data cache
  asm volatile ("mov r0, #0");
  asm volatile ("mcr p15, 0, r0, c7, c7, 0");     // flush v3/v4 cache
  asm volatile ("mcr p15, 0, r0, c8, c7, 0");     // flush v4 TLB

  // Wait a bit.
  for(i=0 ; i<100000 ; i++);


  // Copy the kernel to memory starting at 0xA1000000
  // In reality we want to copy the image to 0xA0008000, but that location
  // holds a second level page table. So if we try to copy directly to the
  // final destination, the system crashes due to a corrupted second level
  // page table. So we temporarily copy it way out into 0xA1000000 first.
  // 
  // Assumption: 0xA1000000 is way out and does not trample on either
  // the code or kernel data.
  for(i=0 ; i<NUM_KERNEL_PAGES*PAGE_SIZE ; i+=4) {
    *((unsigned long *)(rambase+0x01000000+i))=*((unsigned long *)(krnlSrc+i));
  }

  write_atags(mach, rambase, cmdline);

  // FIXME: We arbitrarily add 0x60 to pc and copy 512 bytes starting
  // from there. So if any code is modified between <CRITICAL>
  // and </CRITICAL>, the offset needs to change.
  // <CRITICAL>
  asm volatile ("add %0, pc, #0x60" : "=r" (j));
  for(i=0 ; i<512 ; i++) {
    *((char *)(rambase+0x00001000+i))=*((char *)(i+j));
  }


  asm volatile ("mov lr,   %0;"
      "add pc, lr, #0x00001000;" : : "r" (rambase));

  // </CRITICAL>

  asm volatile ("nop");
  asm volatile ("nop");
  asm volatile ("nop");
  asm volatile ("nop");
  asm volatile ("nop");
  asm volatile ("nop");
  asm volatile ("nop");
  asm volatile ("nop");
  asm volatile ("nop");
  asm volatile ("nop");

  // Let's flush Instruction and Data cache
  asm volatile ("mov r0, #0");
  asm volatile ("mcr p15, 0, r0, c7, c7, 0");     // flush v3/v4 cache
  asm volatile ("mcr p15, 0, r0, c8, c7, 0");     // flush v4 TLB

  // Wait a bit.
  for(i=0 ; i<100000 ; i++);
  
  // Let's disable the MMU and disable data cache.
  asm volatile ("mrc p15, 0, r0, c1, c0, 0");
  asm volatile ("bic r0, r0, #0x00002300");     // clear bits 13, 9:8 (--V- --RS)
  asm volatile ("bic r0, r0, #0x00000087");     // clear bits 7, 2:0 (B--- -CAM)
  asm volatile ("orr r0, r0, #0x00001000");     // set bit 12 (I) I-Cache
  asm volatile ("mcr p15, 0, r0, c1, c0, 0");
  
  // invalidate TLB
  asm volatile ("mov r0, #0");
  asm volatile ("mcr p15, 0, r0, c8, c7, 0");

  // Wait a bit.
  for(i=0 ; i<100000 ; i++);
  
  
  // Actually copy the kernel to 0xA0008000.
  for(i=0 ; i<NUM_KERNEL_PAGES*PAGE_SIZE ; i+=4) {
    *((unsigned long *)(rambase + 0x0008000+i))=*((unsigned long*)(rambase+0x01000000+i));
  }

  // Copy initrd to 0xA0400000.
  for(i=0 ; i<SIZE_INITRD ; i+=4) {
    *((unsigned long *)(rambase + 0x0400000+i))=*((unsigned long*)(rambase+0x01000000+SIZE_KERNEL+i));
  }


  // Wait a bit.
  for(i=0 ; i<10000000 ; i++);

  // Initializing before boot
  asm volatile ("mov r0,   #0");
  asm volatile ("mov r1, %0" : : "r" (mach));     /* mach */ 
  asm volatile ("mov r2,   %0": : "r" (rambase));
  asm volatile ("add r2, r2, #0x00000100");      /* location of ATAGs */


  // Jump to the kernel.
  asm volatile ("mov r3,   #0x00008000");
  asm volatile ("orr r3, r3, %0": : "r" (rambase));
  asm volatile ("mov pc, r3");

  // We should never get here... Just to make the compiler happy.
  return 0;
}


/**
 * Return the start of RAM for the given machine
 */
unsigned long get_rambase(unsigned long who)
{
  switch (who) {
    /* OMAP based */
    case OMAP_PALMTE:   return 0x10000000;      /* Tungsten E */
    case OMAP_PALMTT2:  return 0x10000000;      /* Tungsten T2 */
    
    /* XScale based */
    case T3XSCALE:    return 0xA0000000;      /* Tungsten T3 */
    case XSCALE_PALMLD: return 0xA0000000;      /* LifeDrive */
  }

  return 0;
}

/**
 * Return the cpu type for the given machine
 */
unsigned long get_cpu(unsigned long who)
{
  switch (who) {
    /* OMAP based */
    case OMAP_PALMTE:  /* Tungsten E */
    case OMAP_PALMTT2:  /* Tungsten T2 */
      return OMAP;
          
    /* XScale based */
    case T3XSCALE:    /* Tungsten T3 */
    case XSCALE_PALMLD: /* LifeDrive */
      return XSCALE;
  }

  return 0;
}

/**
 * Return the amount of RAM in the given machine (for pasing to Linux)
 * FIXME: this should be detected somehow
 */
unsigned long get_ram_size(unsigned long who)
{
  switch (who) {
    /* OMAP based */
    case OMAP_PALMTE:   return 0x02000000;      /* Tungsten E (32M) */
    case OMAP_PALMTT2:  return 0x02000000;      /* Tungsten T2 (32M) */
    
    /* XScale based */
    case T3XSCALE:    return 0x04000000;      /* Tungsten T3 (64M) */
    case XSCALE_PALMLD: return 0x02000000;      /* LifeDrive */
  }

  return 0;
}

/**
 * Machine specific initialization (for turning of PalmOS DMA etc)
 */
 
void init_tt2(unsigned long mach)
{
  //io ports settings for uart3
  *((unsigned long*)0xfffe1044)=0xc9fbffff;
  *((unsigned long*)0xfffce004)=0x00003a0e;
  
  //Reset for disable dma LCD
  //irq reset
  *((unsigned long*)0xfffecb00)=0x0;
  *((unsigned long*)0xfffecb04)=0xffffffff;
  //dmaGlobal reset
  *((unsigned short*)0xfffedb00)=0x0;
  *((unsigned short*)0xfffedc00)=0x8;
}


void init_xscale(unsigned long mach)
{ 
  // shut off devices to prevent side-effects of DMA corruption
  // these will be re-enabled by the kernel drivers.
  
  LCCR0 &= ~LCCR0_ENB;    // shut off LCD controller
  //GCR |= ~GCR_ACLINK_OFF;   // shut off the AC97 audio controlller
  
  // reset IRQs in the hope of fixing the bitops.h crash.
  // from kernel26/arch/arm/mach-pxa/irq.c
  
  /* disable all IRQs */
  ICMR = 0;

  /* all IRQs are IRQ, not FIQ */
  ICLR = 0;

  /* clear all GPIO edge detects */
  GFER0 = 0;
  GFER1 = 0;
  GFER2 = 0;
  GRER0 = 0;
  GRER1 = 0;
  GRER2 = 0;
  GEDR0 = GEDR0;
  GEDR1 = GEDR1;
  GEDR2 = GEDR2;
  
  // this does not seem to be required, but if problems occur
  // add it in.
    
  if(mach == XSCALE_PALMLD) { // pxa27x CPU
    /* And similarly for the extra regs on the PXA27x */
    ICMR2 = 0;
    //ICLR2 = 0;
    //GFER3 = 0;
    //GRER3 = 0;
    //GEDR3 = GEDR3;    
  }
}

void init_omap(unsigned long mach)
{
  // todo: do we need anything here?

}


/**
 * Set the ATAGs.
 *
 * <Alex> I have moved this into a seperate function so as to keep the number of
 * local variables in bootLinux to a minimum, meaning the compile can use
 * registers for all of them.
 */
#define SET_TAG_WORD(off, value) *((unsigned long *)((rambase + 0x100)+((off)<<2)))=value
#define SET_TAG_WORD_BYTE(off, byte, value) *((char *)((rambase + 0x100)+((off)<<2)+(byte)))=value

void write_atags(unsigned long mach, unsigned long rambase, char *cmdline)
{
  unsigned long i=0;
  int byte = 0;
  char c;

  // ATAG_CORE
  SET_TAG_WORD(i++,0x00000002);
  SET_TAG_WORD(i++,0x54410001);


  // ATAG_MEM
  SET_TAG_WORD(i++,0x00000004);
  SET_TAG_WORD(i++,0x54410002);
  SET_TAG_WORD(i++,get_ram_size(mach)); // size
  SET_TAG_WORD(i++,rambase); // start

  
  // ATAG_CMDLINE
  for (byte=0; cmdline[byte]; byte++);
  byte = (byte+3) & ~3;
  SET_TAG_WORD(i++,0x00000002+byte/4);
  SET_TAG_WORD(i++,0x54410009);
  byte = 0;
  do {
    c = *cmdline++;
    SET_TAG_WORD_BYTE(i,byte,c);
    if (++byte==4) { byte=0; i++; }
  } while (c);
  while (byte>0) {
    SET_TAG_WORD_BYTE(i,byte,0);
    if (++byte==4) { byte=0; i++; }
  }

  if (get_cpu(mach) == OMAP) {
    // ATAG_OMAP
    SET_TAG_WORD(i++,0x00000005);
    SET_TAG_WORD(i++,0x414f4d50);
    SET_TAG_WORD(i++,0x4f030005);
    SET_TAG_WORD(i++,0x030001c2);
    SET_TAG_WORD(i++,0x00000000);
  }

  // ATAG_NONE
  SET_TAG_WORD(i++,0x00000000);
  SET_TAG_WORD(i++,0x00000000);

}
