/*
 * garux.c
 *
 * Linux Loader for the Palm OS.
 *
 * Copyright (C) 2004 Vinayak R. Borkar
 * Author: Vinayak R. Borkar [vinayakb@users.sourceforge.net]
 *
 * Original version by Romain Goyet [r.goyet@gmail.com]
 *
 * 11-Sep-2005, Alex Osborne [bobofdoom@gmail.com]:
 *   Updated to work with multiple handhelds.
 *
 *   Device support based on the work of:
 *     Nikolay Petukhov   - Palm Tungsten T2
 *     Vladimir Pouzanov  - Palm Tungsten T3
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <PalmOS.h>
#include <PceNativeCall.h>
#include <MemGlue.h>
#include <MemoryMgr.h>
#include <string.h>
#include <stdio.h>

//#include <ByteOrderUtils.h>
#define EndianSwap32(n) (((((unsigned long) n) << 24) & 0xFF000000) |   ((((unsigned long) n) <<  8) & 0x00FF0000) |   ((((unsigned long) n) >>  8) & 0x0000FF00) |   ((((unsigned long) n) >> 24) & 0x000000FF))

#include "garux.h"

// Display the startup warning?
#define WARNING

// Open the serial port for kernel messages
//#define SERIAL
//#define BAUD 9600

char *get_mach_name(int who);
int get_mach();
int main_event_loop();
int alt_event_loop();
int run (void);


FieldType *kernelParametersField;
int mach=UNDEFINED;


int run (void) {
	int i = 0;
	Err retcode;
	UInt32 creator=3007;
	UInt16 featureNum=35;
	UInt32 offset=0;
	MemPtr kernel;
	void *kernelPointer;
	MemHandle asmloaderHandle;
	void *asmloader;
	MemHandle pageHandle;
	void *page;
	UInt32 who = (UInt32)mach;
	//UInt32 v,y;


	//MemGluePtrNew() on palm tt2 allocate <750Kb,FtrPtrNew() allocate more

	retcode = FtrPtrNew (creator,featureNum,0x200000,(void**)&kernel);

	if (retcode != 0) {
		printf ("Ftr not allocate memory\n");
		SysTaskDelay( 1*SysTicksPerSecond() );
		return -1;
	}

	// Write the machine number into the allocated buffer
	DmWrite(kernel, offset, &who, 4);
	offset += 4;
	
	// Fetch kernel command line
	DmWrite(kernel, offset, FldGetTextPtr(kernelParametersField), 256);
	FldSetTextHandle(kernelParametersField, 0);
	offset += 256;

	// Let's close the form now that we don't need the text field anymore
	FrmCloseAllForms();

	// Grab a handle to the assembly loader
	asmloaderHandle = DmGetResource('armc', 0);
	asmloader = MemHandleLock(asmloaderHandle);
	//SysTaskDelay( 1*SysTicksPerSecond() );

	printf("Machine: %s (%d)\n", get_mach_name(mach), mach);
	printf("Loading kernel... at 0x%lx\n",(unsigned long)kernel);

	//SysTaskDelay( 1*SysTicksPerSecond() );

	// Copy the kernel pages into the contiguous memory we allocated.
	kernelPointer = kernel;
	printf("Joining kernel chunks");
	for (i = 0; i < NUM_KERNEL_PAGES; i++) {
		pageHandle = DmGetResource('page', i);
		page = MemHandleLock(pageHandle);

		DmWrite(kernelPointer, offset, page, PAGE_SIZE);
		offset += PAGE_SIZE;

		MemHandleUnlock(pageHandle);
		DmReleaseResource(pageHandle);
		putchar('.');
	}
	printf("OK\n");

	printf("Running assembly loader...\n");
	printf("Ret: %lx\n", PceNativeCall(asmloader, kernel));

	/*	Print out the memory mapping
	for(v=0x4000; v<0x13FFF; v+=4) {
		y=EndianSwap32(*(UInt32*)(v));
		if(y & 0x3) { // mapped?
			if((y & 1) == 1) { // 2nd level
				continue;
				if(y&2) { // fine
					printf("2LF %lx => %lx\n", (v-0x4000) << 18, y >> 12);
				} else { // coarse
					printf("2LC %lx => %lx\n", (v-0x4000) << 18, y >> 10);
				}
			} else { // 1st level
				printf("%lx => %lx\n", (v-0x4000) << 18, y & 0xFFFFF400);
			}
			
			SysTaskDelay( SysTicksPerSecond() / 16 );
		}
	}
	*/
	
	SysTaskDelay( 3*SysTicksPerSecond() );

	// The following code shall not be executed
	FtrPtrFree(creator,featureNum);
	MemHandleUnlock (asmloaderHandle);
	return 0;
}


/**
 * Return the ARM machine number of this handheld.
 * See: http://arm.linux.org.uk/developer/machines/
 */
int get_mach()
{
	UInt32 deviceID;

	FtrGet(sysFtrCreator, sysFtrNumOEMDeviceID, &deviceID);

	switch (deviceID) {
		case 'Cct1': return OMAP_PALMTE;			/* Tungsten E */
		case 'Frg2': return OMAP_PALMTT2;			/* Tungsten T2 */
		case 'Arz1': return T3XSCALE;				/* Tungsten T3 */
		case 'TunX': return XSCALE_PALMLD;			/* LifeDrive */
	}

	return UNDEFINED;
}

/**
 * Return the machine name corresponding to a given mach number.
 */
char *get_mach_name(int who)
{
	switch (who) {
		case OMAP_PALMTE:	return "Palm Tungsten E";
		case OMAP_PALMTT2:	return "Palm Tungsten T2";
		case T3XSCALE:  	return "Palm Tungsten T3";
		case XSCALE_PALMLD: return "Palm LifeDrive";
	}
	
	return "Unknown";
}

int main_event_loop()
{
    UInt16 error_code;
    EventType event;

#ifdef SERIAL
	UInt16 portid=0, irportid=0;
	Err err;
#endif
	
	while (1) {
		EvtGetEvent(&event, 100);
		if (SysHandleEvent(&event))
			continue;
		if (MenuHandleEvent(0, &event, &error_code))
			continue;
		if (event.eType == ctlSelectEvent) {
			if (event.data.ctlSelect.controlID == PalmOSButton) {
				FldSetTextHandle(kernelParametersField, 0);
				FrmCloseAllForms();
#ifdef SERIAL
				SrmClose(portid);
#endif
				return 0;
			} else if (event.data.ctlSelect.controlID == LinuxButton) {
				run();
				return 0;
			}
		}
		if (event.eType == frmLoadEvent) {
			FrmSetActiveForm(FrmInitForm(event.data.frmLoad.formID));
			continue;
		}
		if (event.eType == frmOpenEvent) {
			FrmDrawForm(FrmGetActiveForm());

#ifdef WARNING
			// Initial warning
			if (FrmAlert(StartupAlert) == 1) {
				FrmCloseAllForms();
				return 0;
			}
#endif

			// Initializing boot parameters
			kernelParametersField = (FieldType *)FrmGetObjectPtr(FrmGetFormPtr(MainForm), FrmGetObjectIndex(FrmGetFormPtr(MainForm), KernelParam));

			FldSetTextHandle(kernelParametersField, DmGetResource(strRsc, kernelDefaultParam));
			FldRecalculateField(kernelParametersField, true);

#ifdef SERIAL
			SrmOpen(serPortIrPort, BAUD, &irportid);
			SrmOpen(serPortCradleRS232Port, BAUD, &portid);
			printf("Serial open, port id: %x\n", portid);
			SrmSend(portid, "Garux: PalmOS opened port.\n", 27, &err);
#endif
			continue;
		}
		if (event.eType == appStopEvent) {
			FldSetTextHandle(kernelParametersField, 0);
			FrmCloseAllForms();
#ifdef SERIAL
			SrmClose(portid);
#endif
			return 0;	
		}
		if (FrmGetActiveForm())
			FrmHandleEvent(FrmGetActiveForm(), &event);
	}
	
}

/**
 * Event loop for the bad device form.
 */
int alt_event_loop()
{
    EventType event;
	while (1) {
		EvtGetEvent(&event, 100);
		SysHandleEvent(&event);
		if (event.eType == ctlSelectEvent)
			if (event.data.ctlSelect.controlID == PalmOSButton) {
				FrmCloseAllForms();
				return 0;
			}
			
		if (event.eType == frmLoadEvent) {
			FrmSetActiveForm(FrmInitForm(event.data.frmLoad.formID));
			continue;
		}
		
		if (event.eType == frmOpenEvent) {
			FrmDrawForm(FrmGetActiveForm());
			continue;
		}
		if (event.eType == appStopEvent) {
			FrmCloseAllForms();
			return 0;	
		}
		if (FrmGetActiveForm())
			FrmHandleEvent(FrmGetActiveForm(), &event); 
	}
}


UInt32 PilotMain (UInt16 cmd, void *params, UInt16 flags) {
    if (cmd == sysAppLaunchCmdNormalLaunch) {
		mach = get_mach();

        if (mach != UNDEFINED) {
			// Garux is being run on a supported device
			FrmGotoForm(MainForm);
			return main_event_loop();
			
		} else {
			// Garux is not being run on a supported device. We'll warn, and exit.
			FrmGotoForm(AltForm);
			return alt_event_loop();
		}
	}
	return 0;
}
